/* andnotpostlist.h: Return items which are in A, unless they're in B
 *
 * ----START-LICENCE----
 * Copyright 1999,2000,2001 BrightStation PLC
 * Copyright 2002 Ananova Ltd
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * -----END-LICENCE-----
 */

#ifndef OM_HGUARD_ANDNOTPOSTLIST_H
#define OM_HGUARD_ANDNOTPOSTLIST_H

#include "database.h"
#include "branchpostlist.h"
#include <algorithm>

/** A postlist generated by taking one postlist (the left-hand postlist),
 *  and removing any documents which are in the other (right-hand) postlist.
 *
 *  The weight for a posting is the weight in the left-hand postlist.
 */
class AndNotPostList : public BranchPostList {
    private:
        om_docid lhead, rhead;

	/// Number of documents in the database this postlist is across
	om_doccount dbsize;

        PostList *advance_to_next_match(om_weight w_min, PostList *ret);
    public:
	om_doccount get_termfreq_max() const;
	om_doccount get_termfreq_min() const;
	om_doccount get_termfreq_est() const;

	om_docid  get_docid() const;
	om_weight get_weight() const;
	om_weight get_maxweight() const;

        om_weight recalc_maxweight();

	PostList *next(om_weight w_min);
	PostList *skip_to(om_docid did, om_weight w_min);
	bool   at_end() const;

	std::string get_description() const;

	/** Return the document length of the document the current term
	 *  comes from.
	 */
	virtual om_doclength get_doclength() const;

        AndNotPostList(PostList *left,
		       PostList *right,
		       MultiMatch *matcher_,
		       om_doccount dbsize_);

        PostList *sync_and_skip_to(om_docid id,
				   om_weight w_min,
				   om_docid lh,
				   om_docid rh);
};

inline om_doccount
AndNotPostList::get_termfreq_max() const
{
    DEBUGCALL(MATCH, om_doccount, "AndNotPostList::get_termfreq_max", "");
    // Max is when as many docs as possible on left, and none excluded.
    RETURN(l->get_termfreq_max());
}

inline om_doccount
AndNotPostList::get_termfreq_min() const
{
    DEBUGCALL(MATCH, om_doccount, "AndNotPostList::get_termfreq_min", "");
    // Min is when as few docs as possible on left, and maximum are excluded.
    om_doccount l_min = l->get_termfreq_min();
    om_doccount r_max = r->get_termfreq_max();
    if (l_min > r_max) RETURN(l_min - r_max);
    RETURN(0u);
}

inline om_doccount
AndNotPostList::get_termfreq_est() const
{
    DEBUGCALL(MATCH, om_doccount, "AndNotPostList::get_termfreq_est", "");
    // Estimate assuming independence:
    // P(l and r) = P(l) . P(r)
    // P(l not r) = P(l) - P(l and r) = P(l) . ( 1 - P(r))
    om_doccount est = static_cast<om_doccount>
	    (l->get_termfreq_est() *
	     (1.0 - static_cast<double>(r->get_termfreq_est()) / dbsize));

    RETURN(est);
}

inline om_docid
AndNotPostList::get_docid() const
{
    DEBUGCALL(MATCH, om_docid, "AndNotPostList::get_docid", "");
    RETURN(lhead);
}

// only called if we are doing a probabilistic AND NOT
inline om_weight
AndNotPostList::get_weight() const
{
    DEBUGCALL(MATCH, om_weight, "AndNotPostList::get_weight", "");
    RETURN(l->get_weight());
}

// only called if we are doing a probabilistic AND NOT
inline om_weight
AndNotPostList::get_maxweight() const
{
    DEBUGCALL(MATCH, om_weight, "AndNotPostList::get_maxweight", "");
    RETURN(l->get_maxweight());
}

inline om_weight
AndNotPostList::recalc_maxweight()
{
    DEBUGCALL(MATCH, om_weight, "AndNotPostList::recalc_maxweight", "");
    RETURN(l->recalc_maxweight());
}

inline bool
AndNotPostList::at_end() const
{
    DEBUGCALL(MATCH, bool, "AndNotPostList::at_end", "");
    RETURN(lhead == 0);
}

inline std::string
AndNotPostList::get_description() const
{
    return "(" + l->get_description() + " AndNot " + r->get_description() + ")";
}

inline om_doclength
AndNotPostList::get_doclength() const
{
    DEBUGCALL(MATCH, om_doclength, "AndNotPostList::get_doclength", "");
    RETURN(l->get_doclength());
}

#endif /* OM_HGUARD_ANDNOTPOSTLIST_H */
